<?
  /*
  * 環境クラス(2012/04/07)
      @author   TagajoTown
  */
  
  /*
    トップレベル環境クラスを返す
  */
  function get_global_environment()
  {
    $env = new Environment();
    $env->add( '+' ,       create_function( '$a, $b',  'return $a + $b;' ) );
    $env->add( '-' ,       create_function( '$a, $b',  'return $a - $b;' ) );
    $env->add( '*' ,       create_function( '$a, $b',  'return $a * $b;' ) );
    $env->add( '/' ,       create_function( '$a, $b',  'return $a / $b;' ) );
    $env->add( 'not' ,     create_function( '$a',      'return !$a;' ) );
    $env->add( '>' ,       create_function( '$a, $b',  'return $a > $b;' ) );
    $env->add( '<' ,       create_function( '$a, $b',  'return $a < $b;' ) );
    $env->add( '>=' ,      create_function( '$a, $b',  'return $a >= $b;' ) );
    $env->add( '<=' ,      create_function( '$a, $b',  'return $a <= $b;' ) );
    $env->add( '=' ,       create_function( '$a, $b',  'return $a == $b;' ) );
    $env->add( 'equal?' ,  create_function( '$a, $b',  'return $a == $b;' ) );
    $env->add( 'eq?' ,     create_function( '$a, $b',  'return $a == $b;' ) );
    $env->add( 'length' ,  create_function( '$a',      'return count( $a );' ) );
    $env->add( 'cons' ,    create_function( '$a, $b',  'array_unshift( $a, $b ); return $a;' ) );
    $env->add( 'car' ,     create_function( '$a',      'return $a[ 0 ];' ) );
    $env->add( 'cdr' ,     create_function( '$a',      'return array_slice( $a, 1, count( $a ) );' ) );
    $env->add( 'append' ,  create_function( '$a, $b',  'return array_merge( $a, $b );' ) );
    $env->add( 'list' ,    create_function( '',        'return func_get_args();' ) );
    $env->add( 'list?' ,   create_function( '$a',      'return is_array( $a );' ) );
    $env->add( 'null?' ,   create_function( '$a',      'return count( $a ) == 0 ? true : false;' ) );
    $env->add( 'symbol?' , create_function( '$a',      'return is_string( $a );' ) );
    
    return $env;
  }
  
  
  /**
    環境クラス
  */
  class Environment
  {
    private     $outer;
    private     $symbols;
    
    /**
      環境クラスコンストラクタ
      @param $outer 外側のスコープ環境
    */
    public function __construct( &$outer = null )
    {
      $this->symbols = array();
      $this->outer = &$outer;
    }
    
    /**
      環境へシンボル登録。
      2つの配列をarray_combineで合成し、環境へ登録
      @param $symbols Symbolオブジェクト配列（シンボル名を保持）
      @param $values  シンボルの値の配列
    */
    public function zip( $symbols, $values )
    {
      // $symbolsはSymbolオブジェクト。
      $keys = array();
      foreach( $symbols AS $key ) {
        $keys[] = $key->value;
      }
      
      $temp = array_combine( $keys, $values );
      $this->symbols = array_merge( $this->symbols, $temp );
    }
    
    /**
      環境へシンボル登録。
      値2つを取り、環境へ登録。同じ名前のときは上書きされる。
      @param $name  シンボル名
      @param $value シンボルの値
    */
    public function add( $name, $value )
    {
      $this->symbols[ $name ] = $value;
    }
    
    /**
      シンボルの値を取得する。
      見つからなかった場合は「UnboundException」をスローする
      @param $name シンボル名
      @return 値
      @exception UnboundException
    */
    public function get( $name )
    {
      if( $this->exists( $name ) ) {
        if( isSet( $this->symbols[ $name ] ) ) {
          return $this->symbols[ $name ];
        } else {
          return $this->outer->get( $name );
        }
      } else {
        throw new UnboundException( sprintf( "[%s] is Unbound.\n", $name ), $this );
      }
    }
    
    /**
      シンボルが存在するかを確認する。
      @param $name シンボル名
      @return boolean
    */
    public function exists( $name )
    {
      if( isSet( $this->symbols[ $name ] ) ) {
        return true;
      } else {
        if( $this->outer == null ) {
          return false;
        } else {
          return $this->outer->exists( $name );
        }
      }
    }
    
  }
  
?>
