<?
  /**
    評価器クラス(2012/04/07)
    @author   TagajoTown
  */
  
  class Evaluator
  {
    /**
      定数、シンボルまたはシンボルの配列を評価する。再帰関数。
      @param $symbols 定数、シンボル、またはシンボルの配列
      @param $env     環境オブジェクト。指定されない場合はトップレベル環境。
    */
    public static function lisp_eval( $symbols, $env = null )
    {
      // 環境が指定されていないときはトップレベルの環境を指定
      if( $env == null ) {
        global $global_environment;
        $env = $global_environment;
      }
      
      // シンボルの場合
      if( is_a( $symbols, 'Symbol' ) ) {
        return $env->get( $symbols->value );
      
      // 定数の場合
      } elseif( is_scalar( $symbols ) ) {
        return $symbols;
      
      // 以降リストの場合
      } elseif( $symbols[ 0 ]->value == 'quote' ) {     // (quote exp)
        list( $quote, $exp ) = $symbols;
        return $exp;
        
      } elseif( $symbols[ 0 ]->value == 'if' ) {        // (if test conseq alt)
        list( $if, $test, $conseq, $alt ) = $symbols;
        if( self::lisp_eval( $test, $env ) ) {
          return self::lisp_eval( $conseq, $env );
        } else {
          return self::lisp_eval( $alt, $env );
        }
        
      } elseif( $symbols[ 0 ]->value == 'define' ) {      // (define var (lambda exp))
        list( $define, $var, $exp ) = $symbols;
        $env->add( $var->value, self::lisp_eval( $exp, $env ) );
        
      } elseif( $symbols[ 0 ]->value == 'lambda' ) {      // (lambda vars exp)

        list( $lambda, $vars, $exp ) = $symbols;
        $str_lambda  = '$args = func_get_args();' . "\n";
        $str_lambda .= '$env  = array_pop( $args );' . "\n";
        $str_lambda .= '$vars = unserialize( stripslashes( ' . "'" . addslashes( serialize( $vars ) ) . "'" . ' ) ); ' . "\n";
        $str_lambda .= '$exp  = unserialize( stripslashes( ' . "'" . addslashes( serialize( $exp ) ) . "'" . ' ) ); ' . "\n";
        $str_lambda .= '$new_env = new Environment( &$env );' . "\n";
        $str_lambda .= '$new_env->zip( $vars, $args );' . "\n";
        $str_lambda .= 'return Evaluator::lisp_eval( $exp, $new_env );' . "\n";

        $lambda_function = create_function( '', $str_lambda );
        return $lambda_function;
      
      } elseif( $symbols[ 0 ]->value == 'begin' ) {       // (begin exps)
        each( $symbols );
        foreach( $symbols AS $symbol ) {
          $val = self::lisp_eval( $symbol->value, $env );
        }
        return $val;
        
      } else {                                            // 関数評価
        $exps = array();
        
        foreach( $symbols AS $symbol ) {
          $exps[] = self::lisp_eval( $symbol, $env );
        }
        $proc = array_shift( $exps );
        
        /*
          小細工
          関数が使う環境オブジェクトを関数の引数として渡す。
        */
        array_push( $exps, $env );
        $ret  = call_user_func_array( $proc, $exps );
        return $ret;
      }
    }
  }
?>
